/*
 * See Licensing and Copyright notice in naev.h
 */
#pragma once

/* Forward declarations. */
typedef struct StarSystem_ StarSystem;
typedef struct Spob_       Spob;
typedef struct JumpPoint_  JumpPoint;
